"""Routines to evaluate the correlation between HPGe and SiPM signals."""

from __future__ import annotations

from collections.abc import Sequence

import awkward as ak
import numpy as np
import scipy.stats
from numpy.typing import ArrayLike


def l200_combined_test_stat(
    t0: ak.Array,
    amp: ak.Array,
    geds_t0: ak.Array,
    ts_bkg_prob: float,
    rc_density: Sequence[float],
) -> ak.Array:
    """Combined L200 LAr veto classifier.

    Where combined means taking channel-specific parameters into account.

    `t0` and `amp` must be in the format of a 3-dimensional Awkward array,
    where the innermost dimension labels the SiPM pulse, the second one labels
    the SiPM channel and the outermost one labels the event.

    Parameters
    ----------
    t0
        arrival times of pulses in ns, split by channel.
    amp
        amplitude of pulses in p.e., split by channel.
    geds_t0
        t0 (ns) of the HPGe signal.
    ts_bkg_prob
        probability for a pulse coming from some uncorrelated physics (uniform
        distribution). needed for the LAr scintillation time pdf.
    rc_density
        density array of the random coincidence LAr energy distribution (total
        energy summed over all channels, in p.e.). Derived from forced trigger
        data.
    """
    # flatten the data in the last axis (i.e. merge all channels together)
    # TODO: implement channel distinction
    t0 = ak.flatten(t0, axis=-1)
    amp = ak.flatten(amp, axis=-1)

    # subtract the HPGe t0 from the SiPM pulse t0s
    rel_t0 = t0 - geds_t0

    return l200_test_stat(rel_t0, amp, ts_bkg_prob, rc_density)


def l200_test_stat(relative_t0, amp, ts_bkg_prob, rc_density):
    """Compute the test statistics.

    Parameters
    ----------
    relative_t0
        t0 (ns) of the SiPM pulses relative to the HPGe t0.
    amp
        amplitude in p.e. of the SiPM pulses.
    """
    # convert to integer number of pes
    n_pes = pulse_amp_round(amp)

    # compute total number of pes in event
    n_pe_tot = np.sum(n_pes, axis=-1)
    # if no pes in the event, use nan instead of zero (because of division later)
    n_pe_tot = np.where(n_pe_tot == 0, np.nan, n_pe_tot)

    # calculate the test statistic term related to the time distribution
    ts_time = -ak.sum(
        ak.transform(
            lambda layouts, **kwargs: _ak_l200_test_stat_time_term(
                layouts, ts_bkg_prob, **kwargs
            ),
            relative_t0,
            amp,
        ),
        axis=-1,
    )
    # calculate the amplitude contribution
    ts_amp = [l200_rc_amp_logpdf(n, rc_density) for n in n_pe_tot]

    # for events with no light, set the test statistic value to +inf
    t_stat = np.where(np.isnan(n_pe_tot), np.inf, ts_time / n_pe_tot + ts_amp)

    return t_stat


# need to define this function and use it with ak.transform() because scipy
# routines are not NumPy universal functions
def _ak_l200_test_stat_time_term(layouts, ts_bkg_prob, **kwargs):
    """Awkward transform to compute the per-pulse terms of the test statistics.

    The two arguments are the pulse times `t0` relative to the HPGe trigger and
    their amplitude `amp`. The function has to be invoked as
    ``ak.transform(_ak_l200_test_stat_terms, t0, amp,
    ...)``.
    """
    # sanity check
    assert len(layouts) == 2

    if not all([lay.is_numpy for lay in layouts]):
        return

    # these are the two supported arguments
    t0 = layouts[0].data
    amp = layouts[1].data

    # sanity check
    assert len(t0) == len(amp)

    # convert to integer number of pes
    n_pes = pulse_amp_round(amp)

    # calculate the time term of the test statistic
    ts_time = n_pes * np.log(l200_tc_time_pdf(t0, bkg_prob=ts_bkg_prob))

    return ak.contents.NumpyArray(ts_time)


def pulse_amp_round(amp: float | ArrayLike):
    """Get the most likely (integer) number of photo-electrons."""
    # promote all amps < 1 to 1. standard rounding to nearest for amps > 1
    return ak.where(amp < 1, np.ceil(amp), np.floor(amp + 0.5))


def l200_tc_time_pdf(
    t0: float | ArrayLike,
    *,
    domain_ns: tuple[float] = (-1_000, 5_000),
    tau_singlet_ns: float = 6,
    tau_triplet_ns: float = 1100,
    sing2tot_ratio: float = 1 / 3,
    t0_res_ns: float = 35,
    t0_bias_ns: float = -80,
    bkg_prob: float = 0.42,
) -> float | ArrayLike:
    """The L200 experimental LAr scintillation pdf

    The theoretical scintillation pdf convoluted with a Normal distribution
    (experimental effects) and summed to a uniform distribution (uncorrelated
    pulses).

    This routine does not work with :class:`ak.Array`, since SciPy functions
    are not universal. See :func:`_ak_l200_test_stat_time_term` for an example
    Awkward transform that does the job.

    Parameters
    ----------
    t0
        arrival times of the SiPM pulses in ns.
    tau_singlet_ns
        The lifetime of the LAr singlet state in ns.
    tau_triplet_ns
        The lifetime of the LAr triplet state in ns.
    sing2tot_ratio
        The singlet-to-total excitation probability ratio.
    t0_res_ns
        sigma (ns) of the normal distribution.
    t0_bias_ns
        mean (ns) of the normal distribution.
    bkg_prob
        probability for a pulse coming from some uncorrelated physics (uniform
        distribution).
    """
    if np.any(t0 > domain_ns[1]) or np.any(t0 < domain_ns[0]):
        msg = f"{t0=} out of bounds for {domain_ns=}"
        raise ValueError(msg)

    return (
        # the triplet
        (1 - bkg_prob)
        * (
            (1 - sing2tot_ratio)
            * scipy.stats.exponnorm.pdf(
                t0, tau_triplet_ns / t0_res_ns, loc=t0_bias_ns, scale=t0_res_ns
            )
            # the singlet
            + sing2tot_ratio
            * scipy.stats.exponnorm.pdf(
                t0, tau_singlet_ns / t0_res_ns, loc=t0_bias_ns, scale=t0_res_ns
            )
        )
        # the random coincidences (uniform pdf)
        + bkg_prob
        * scipy.stats.uniform.pdf(t0, domain_ns[0], domain_ns[1] - domain_ns[0])
    )


def l200_rc_amp_logpdf(
    n_pes,
    rc_density=None,
    logexp_cont_slope=-1 / 10,
):
    """The L200 experimental random coincidence (RC) amplitude pdf

    Parameters
    ----------
    n_pes
        number of photoelectrons.
    rc_density
        density array of the random coincidence LAr energy distribution (total
        energy summed over all channels, in p.e.). Derived from forced trigger
        data.
    logexp_cont_slope
        slope for exponential analytical continuation.

    """
    # analytical continuation must be decaying exponential function.
    assert logexp_cont_slope < 0

    if rc_density is None:
        rc_density = [1]

    # up to 15 p.e., take the experimental values from the density.
    limit = min(len(rc_density) - 1, 15)
    if n_pes <= limit:
        return np.log(rc_density[int(n_pes)])
    # analytical continuation: log of an exponential function.
    else:
        return logexp_cont_slope * (n_pes - int(limit)) + np.log(rc_density[int(limit)])



"""
pygama convenience functions for fitting binned data
"""

import logging
from typing import Callable, Optional

import numpy as np
from iminuit import Minuit, cost

import pygama.math.histogram as pgh
from pygama.math.functions.gauss import nb_gauss_amp

log = logging.getLogger(__name__)


def fit_binned(
    func: Callable,
    hist: np.ndarray,
    bins: np.ndarray,
    var: np.ndarray = None,
    guess: np.ndarray = None,
    cost_func: str = "LL",
    extended: bool = True,
    simplex: bool = False,
    bounds: tuple[tuple[float, float], ...] = None,
    fixed: tuple[int, ...] = None,
) -> tuple[np.ndarray, ...]:
    """
    Do a binned fit to a histogram.

    Default is extended Log Likelihood fit, with option for either Least
    Squares or other cost function.

    Parameters
    ----------
    func
        the function to fit, if using LL as method needs to be a cdf
    hist, bins, var
        histogrammed data
    guess
        initial guess parameters
    cost_func
        cost function to use
    extended
        run extended or non extended fit
    simplex
        whether to include a round of simpson minimisation before main minimisation
    bounds
        list of tuples with bounds can be None, e.g. [(0,None), (0,10)]
    fixed
        list of parameter indices to fix

    Returns
    -------
    coeff
        Returned fit parameters
    error
        Iminuit errors
    cov_matrix
        Covariance matrix
    """
    if guess is None:
        raise NotImplementedError(
            "auto-guessing not yet implemented, you must supply a guess."
        )

    if cost_func == "LL":

        if var is not None:
            t_arr = np.zeros((len(hist), 2))
            t_arr[:, 0] = hist
            t_arr[:, 1] = var
            hist = t_arr

        if extended is True:
            cost_func = cost.ExtendedBinnedNLL(hist, bins, func)

        else:
            cost_func = cost.BinnedNLL(hist, bins, func)

    elif cost_func == "Least Squares":

        if var is None:
            var = hist  # assume Poisson stats if variances are not provided

        if len(bins) == len(hist) + 1:
            bin_centres = pgh.get_bin_centers(bins)
        else:
            bin_centres = bins

        # skip "okay" bins with content 0 +/- 0
        # if bin content is non-zero but var = 0 let the user see the warning
        zeros = hist == 0
        zero_errors = var == 0
        mask = ~(zeros & zero_errors)
        hist = hist[mask]
        var = np.sqrt(var[mask])
        xvals = bin_centres[mask]
        cost_func = cost.LeastSquares(xvals, hist, var, func)

    m = Minuit(cost_func, *guess)
    if bounds is not None:
        if isinstance(bounds, dict):
            for key, val in bounds.items():
                m.limits[key] = val
    if fixed is not None:
        for fix in fixed:
            m.fixed[fix] = True
    if simplex is True:
        m.simplex().migrad()
    else:
        m.migrad()
    m.hesse()
    return m.values, m.errors, m.covariance


def goodness_of_fit(
    hist: np.ndarray,
    bins: np.ndarray,
    var: np.ndarray,
    func: Callable,
    pars: np.ndarray,
    method: str = "var",
    scale_bins: bool = False,
) -> tuple[float, int]:
    """
    Compute chisq and dof of fit

    Parameters
    ----------
    hist, bins, var
        histogram data. var can be None if hist is integer counts
    func
        the function that was fit to the hist
    pars
        the best-fit pars of func. Assumes all pars are free parameters
    method
        Sets the choice of "denominator" in the chi2 sum
        'var': user passes in the variances in var (must not have zeros)
        'Pearson': use func (hist must contain integer counts)
        'Neyman': use hist (hist must contain integer counts and no zeros)

    Returns
    -------
    chisq
        the summed up value of chisquared
    dof
        the number of degrees of freedom
    """
    # arg checks
    if method == "var":
        if var is None:
            raise RuntimeError("var must be non-None to use method 'var'")
        if np.any(var == 0):
            raise ValueError("var cannot contain zeros")
    if method == "Neyman" and np.any(hist == 0):
        raise ValueError("hist cannot contain zeros for Neyman method")

    # compute expected values
    yy = func(pgh.get_bin_centers(bins), *pars)
    if scale_bins is True:
        yy *= pgh.get_bin_widths(bins)

    if method == "LR":
        log_lr = 2 * np.sum(
            np.where(
                hist > 0,
                yy - hist + hist * np.log((hist + 1.0e-99) / (yy + 1.0e-99)),
                yy - hist,
            )
        )
        dof = len(hist) - len(pars)
        return log_lr, dof

    else:
        # compute chi2 numerator and denominator
        numerator = (hist - yy) ** 2
        if method == "var":
            denominator = var
        elif method == "Pearson":
            denominator = yy
        elif method == "Neyman":
            denominator = hist
        else:
            raise NameError(f"goodness_of_fit: unknown method {method}")

        # compute chi2 and dof
        chisq = np.sum(numerator / denominator)
        dof = len(hist) - len(pars)
        return chisq, dof


def poisson_gof(
    pars: np.ndarray,
    func: Callable,
    hist: np.ndarray,
    bins: np.ndarray,
    is_integral: bool = False,
    **kwargs,
) -> float:
    """
    Calculate the goodness of fit for the Poisson likelihood.

    Parameters
    ----------
    pars
        The parameters of the function, func
    func
        The function that was fit
    hist
        The data that were fit
    bins
        The bins of the histogram that was fit to
    is_integral
        Tells get_bin_estimates if the function is an integral function

    Returns
    -------
    Poisson G.O.F.

    Notes
    -----
    The Poisson likelihood does not give a good GOF until the counts are very
    high and all the poisson stats are roughly gaussian and you don't need it
    anyway. But the G.O.F. is calculable for the Poisson likelihood. So we do
    it here.
    """
    mu = pgh.get_bin_estimates(pars, func, bins, is_integral, **kwargs)
    return 2.0 * np.sum(mu + hist * (np.log((hist + 1.0e-99) / (mu + 1.0e-99)) + 1))


def gauss_mode_width_max(
    hist: np.ndarray,
    bins: np.ndarray,
    var: Optional[np.ndarray] = None,
    mode_guess: Optional[float] = None,
    n_bins: Optional[int] = 5,
    cost_func: str = "Least Squares",
    inflate_errors: Optional[bool] = False,
    gof_method: Optional[str] = "var",
) -> tuple[np.ndarray, ...]:
    r"""
    Get the max, mode, and width of a peak based on gauss fit near the max
    Returns the parameters of a gaussian fit over n_bins in the vicinity of the
    maximum of the hist (or the max near mode_guess, if provided). This is
    equivalent to a Taylor expansion around the peak maximum because near its
    maximum a Gaussian can be approximated by a 2nd-order polynomial in x:

    .. math::

        A \exp[ -(x-\mu)^2 / 2 \sigma^2 ]
        \simeq A [ 1 - (x-\mu)^2 / 2 \sigma^2 ]
        = A - (1/2!) (A/\sigma^2) (x-\mu)^2

    The advantage of using a gaussian over a polynomial directly is that the
    gaussian parameters are the ones we care about most for a peak, whereas for
    a poly we would have to extract them after the fit, accounting for
    covariances. The gaussian also better approximates most peaks farther down
    the peak. However, the gauss fit is nonlinear and thus less stable.

    Parameters
    ----------
    hist
        The values of the histogram to be fit
    bins
        The bin edges of the histogram to be fit
    var
        The variances of the histogram values. If not provided, square-root
        variances are assumed.
    mode_guess
        An x-value (not a bin index!) near which a peak is expected. The
        algorithm fits around the maximum within +/- n_bins of the guess. If not
        provided, the center of the max bin of the histogram is used.
    n_bins
        The number of bins (including the max bin) to be used in the fit. Also
        used for searching for a max near mode_guess
    cost_func
        Passed to fit_binned()
    inflate_errors
        If true, the parameter uncertainties are inflated by sqrt(chi2red)
        if it is greater than 1
    gof_method
        method flag for goodness_of_fit

    Returns
    -------
    pars : ndarray containing the parameters (mode, sigma, maximum) of the gaussian fit

        - mode : the estimated x-position of the maximum
        - sigma : the estimated width of the peak. Equivalent to a gaussian
          width (sigma), but based only on the curvature within n_bins of
          the peak.  Note that the Taylor-approxiamted curvature of the
          underlying function in the vicinity of the max is given by max /
          sigma^2
        - maximum : the estimated maximum value of the peak
    cov : 3x3 ndarray of floats
        The covariance matrix for the 3 parameters in pars
    """
    bin_centers = pgh.get_bin_centers(bins)
    if mode_guess is not None:
        i_0 = pgh.find_bin(mode_guess, bins)
    else:
        i_0 = np.argmax(hist)
        mode_guess = bin_centers[i_0]
    amp_guess = hist[i_0]
    i_0 -= int(np.floor(n_bins / 2))
    i_n = i_0 + n_bins
    width_guess = bin_centers[i_n] - bin_centers[i_0]
    vv = None if var is None else var[i_0:i_n]
    guess = (mode_guess, width_guess, amp_guess)
    pars, errors, cov = fit_binned(
        nb_gauss_amp,
        hist[i_0:i_n],
        bins[i_0 : i_n + 1],
        vv,
        guess=guess,
        cost_func=cost_func,
    )
    if pars[1] < 0:
        pars[1] = -pars[1]
    if inflate_errors:
        chi2, dof = goodness_of_fit(hist, bins, var, nb_gauss_amp, pars)
        if chi2 > dof:
            cov *= chi2 / dof
    return np.asarray([pars["mu"], pars["sigma"], pars["a"]]), np.asarray(cov)


def gauss_mode_max(
    hist: np.ndarray, bins: np.ndarray, **kwargs
) -> tuple[np.ndarray, ...]:
    """Alias for gauss_mode_width_max that just returns the max and mode

    See Also
    --------
    gauss_mode_width_max

    Returns
    -------
    pars : ndarray with the parameters (mode, maximum) of the gaussian fit
        mode : the estimated x-position of the maximum
        maximum : the estimated maximum value of the peak
    cov : 2x2 ndarray of floats
        The covariance matrix for the 2 parameters in pars

    Examples
    --------
    >>> import pygama.math.histograms as pgh
    >>> from numpy.random import normal
    >>> import pygama.math.peak_fitting as pgf
    >>> hist, bins, var = pgh.get_hist(normal(size=10000), bins=100, range=(-5,5))
    >>> pgf.gauss_mode_max(hist, bins, var=var, n_bins=20)
    """
    pars, cov = gauss_mode_width_max(hist, bins, **kwargs)
    if pars is None or cov is None:
        raise RuntimeError("fit binned failed to work")
    return pars[::2], cov[::2, ::2]  # skips "sigma" rows and columns


def gauss_mode(hist: np.ndarray, bins: np.ndarray, **kwargs) -> tuple[float, float]:
    """Alias for gauss_mode_max that just returns the mode (position) of a peak

    See Also
    --------
    gauss_mode_max
    gauss_mode_width_max
    Returns
    -------
    mode : the estimated x-position of the maximum
    dmode : the uncertainty in the mode
    """
    pars, cov = gauss_mode_width_max(hist, bins, **kwargs)
    if pars is None or cov is None:
        return None, None
    return pars[0], np.sqrt(cov[0, 0])


def taylor_mode_max(
    hist: np.ndarray,
    bins: np.ndarray,
    var: Optional[np.ndarray] = None,
    mode_guess: Optional[float] = None,
    n_bins: int = 5,
) -> tuple[np.ndarray, ...]:
    """Get the max and mode of a peak based on Taylor exp near the max
    Returns the amplitude and position of a peak based on a poly fit over n_bins
    in the vicinity of the maximum of the hist (or the max near mode_guess, if provided)

    Parameters
    ----------
    hist
        The values of the histogram to be fit. Often: send in a slice around a peak
    bins
        The bin edges of the histogram to be fit
    var
        The variances of the histogram values. If not provided, square-root
        variances are assumed.
    mode_guess
        An x-value (not a bin index!) near which a peak is expected. The
        algorithm fits around the maximum within +/- n_bins of the guess. If not
        provided, the center of the max bin of the histogram is used.
    n_bins
        The number of bins (including the max bin) to be used in the fit. Also
        used for searching for a max near mode_guess

    Returns
    -------
    pars : 2-tuple with the parameters (mode, max) of the fit
        mode : the estimated x-position of the maximum
        maximum : the estimated maximum value of the peak
    cov : 2x2 matrix of floats
        The covariance matrix for the 2 parameters in pars

    Examples
    --------
    >>> import pygama.analysis.histograms as pgh
    >>> from numpy.random import normal
    >>> import pygama.analysis.peak_fitting as pgf
    >>> hist, bins, var = pgh.get_hist(normal(size=10000), bins=100, range=(-5,5))
    >>> pgf.taylor_mode_max(hist, bins, var, n_bins=5)
    """
    if mode_guess is not None:
        i_0 = pgh.find_bin(mode_guess, bins)
    else:
        i_0 = np.argmax(hist)
    i_0 -= int(np.floor(n_bins / 2))
    i_n = i_0 + n_bins
    wts = None if var is None else 1 / np.sqrt(var[i_0:i_n])

    pars, cov = np.polyfit(
        pgh.get_bin_centers(bins)[i_0:i_n], hist[i_0:i_n], 2, w=wts, cov="unscaled"
    )
    mode = -pars[1] / 2 / pars[0]
    maximum = pars[2] - pars[0] * mode**2
    # build the jacobian to compute the output covariance matrix
    jac = np.array(
        [
            [pars[1] / 2 / pars[0] ** 2, -1 / 2 / pars[0], 0],
            [pars[1] ** 2 / 4 / pars[0] ** 2, -pars[1] / 2 / pars[0], 1],
        ]
    )
    cov_jact = np.matmul(cov, jac.transpose())
    cov = np.matmul(jac, cov_jact)
    return (mode, maximum), cov



import numpy as np
from scipy.stats import moyal as scipy_moyal

from pygama.math.functions.moyal import moyal
from pygama.math.functions.pygama_continuous import PygamaContinuous


def test_moyal_pdf():
    x = np.arange(-10, 12)
    mu = 3
    sigma = 3

    y = moyal.pdf(x, mu, sigma)
    y_direct = moyal.get_pdf(x, mu, sigma)
    scipy_y = scipy_moyal.pdf(x, mu, sigma)

    assert isinstance(moyal, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    x_lo = -10
    x_hi = 50

    y_sig, y_ext = moyal.pdf_ext(x, x_lo, x_hi, n_sig, mu, sigma)
    assert np.allclose(y_ext, n_sig * scipy_y, rtol=1e-8)
    assert np.allclose(y_sig, n_sig, rtol=1e-3)

    normalization = np.diff(scipy_moyal.cdf(x[np.array([0, -1])], mu, sigma))
    y_norm = moyal.pdf_norm(x, x[0], x[-1], mu, sigma)

    assert np.allclose(y_norm, scipy_y / normalization, rtol=1e-8)


def test_moyal_cdf():
    x = np.arange(-10, 12)
    mu = 1
    sigma = 20

    y = moyal.cdf(x, mu, sigma)
    y_direct = moyal.get_cdf(x, mu, sigma)
    scipy_y = scipy_moyal.cdf(x, mu, sigma)

    assert isinstance(moyal, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    y_ext = moyal.cdf_ext(x, n_sig, mu, sigma)
    assert np.allclose(y_ext, n_sig * scipy_y, rtol=1e-8)

    normalization = np.diff(scipy_moyal.cdf(x[np.array([0, -1])], mu, sigma))
    y_norm = moyal.cdf_norm(x, x[0], x[-1], mu, sigma)

    assert np.allclose(y_norm, scipy_y / normalization, rtol=1e-8)


def test_required_args():
    names = moyal.required_args()
    assert names[0] == "mu"
    assert names[1] == "sigma"


def test_name():
    assert moyal.name == "moyal"



"""
pygama's command line interface utilities.
"""

import argparse
import logging
import os
import sys

import pygama
import pygama.logging
from pygama.hit import build_hit


def pygama_cli():
    """pygama's command line interface.

    Defines the command line interface (CLI) of the package, which exposes some
    of the most used functions to the console.  This function is added to the
    ``entry_points.console_scripts`` list and defines the ``pygama`` executable
    (see ``setuptools``' documentation). To learn more about the CLI, have a
    look at the help section:

    .. code-block:: console

      $ pygama --help
      $ pygama build-hit --help  # help section for a specific sub-command
    """

    parser = argparse.ArgumentParser(
        prog="pygama", description="pygama's command-line interface"
    )

    # global options
    parser.add_argument(
        "--version", action="store_true", help="""Print pygama version and exit"""
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="""Increase the program verbosity""",
    )
    parser.add_argument(
        "--debug",
        "-d",
        action="store_true",
        help="""Increase the program verbosity to maximum""",
    )

    subparsers = parser.add_subparsers()

    add_build_hit_parser(subparsers)

    if len(sys.argv) < 2:
        parser.print_usage(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    if args.verbose:
        pygama.logging.setup(logging.DEBUG)
    elif args.debug:
        pygama.logging.setup(logging.DEBUG, logging.root)
    else:
        pygama.logging.setup()

    if args.version:
        print(pygama.__version__)  # noqa: T201
        sys.exit()

    args.func(args)


def add_build_hit_parser(subparsers):
    """Configure :func:`.hit.build_hit.build_hit` command line interface"""

    parser_r2d = subparsers.add_parser(
        "build-hit",
        description="""Process LH5 dsp files and produce a hit file using a
                       JSON configuration""",
    )
    parser_r2d.add_argument(
        "dsp_lh5_file",
        nargs="+",
        help="""Input dsp LH5 file. Can be a single file or a list of them""",
    )
    parser_r2d.add_argument(
        "--config",
        "-c",
        required=True,
        help=""""JSON file holding configuration of column operations on DSP output""",
    )
    parser_r2d.add_argument(
        "--hdf5-groups",
        "-g",
        nargs="*",
        default=None,
        help="""Name of group in the LH5 file. By default process all base
                groups. Supports wildcards""",
    )
    parser_r2d.add_argument(
        "--output",
        "-o",
        default=None,
        help="""Name of output file, if only one is supplied. By default,
                output to <input-filename>_hit.lh5""",
    )
    parser_r2d.add_argument(
        "--max-rows",
        "-n",
        default=None,
        type=int,
        help="""Number of rows to process. By default do the whole file""",
    )
    parser_r2d.add_argument(
        "--chunk",
        "-k",
        default=3200,
        type=int,
        help="""Number of waveforms to read from disk at a time. Default is
                3200""",
    )

    group = parser_r2d.add_mutually_exclusive_group()
    group.add_argument(
        "--overwrite",
        "-w",
        action="store_const",
        const="of",
        dest="writemode",
        default="w",
        help="""Overwrite file if it already exists. Default option""",
    )
    group.add_argument(
        "--update",
        "-u",
        action="store_const",
        const="u",
        dest="writemode",
        help="""Update existing file with new values""",
    )
    group.add_argument(
        "--append",
        "-a",
        action="store_const",
        const="a",
        dest="writemode",
        help="""Append values to existing file""",
    )

    parser_r2d.set_defaults(func=build_hit_cli)


def build_hit_cli(args):
    """Passes command line arguments to :func:`.hit.build_hit.build_hit`."""

    if len(args.dsp_lh5_file) > 1 and args.output is not None:
        raise NotImplementedError("not possible to set multiple output file names yet")

    out_files = []
    if len(args.dsp_lh5_file) == 1:
        if args.output is None:
            basename = os.path.splitext(os.path.basename(args.dsp_lh5_file[0]))[0]
            basename = basename.removesuffix("_dsp")
            out_files.append(f"{basename}_hit.lh5")
        else:
            out_files.append(args.output)
    else:
        for file in args.dsp_lh5_file:
            basename = os.path.splitext(os.path.basename(file))[0]
            basename = basename.removesuffix("_dsp")
            out_files.append(f"{basename}_hit.lh5")

    for i in range(len(args.dsp_lh5_file)):
        build_hit(
            args.dsp_lh5_file[i],
            args.config,
            outfile=out_files[i],
            lh5_tables=args.hdf5_groups,
            n_max=args.max_rows,
            wo_mode=args.writemode,
            buffer_len=args.chunk,
        )



import numpy as np
from scipy.stats import crystalball as scipy_crystal_ball

from pygama.math.functions.crystal_ball import crystal_ball
from pygama.math.functions.pygama_continuous import PygamaContinuous


def test_crystalball_pdf():
    x = np.arange(-10, 12)
    beta = 2
    m = 3
    mu = 1
    sigma = 2

    par_array = [mu, sigma, beta, m]

    y = crystal_ball.pdf(x, *par_array)
    y_direct = crystal_ball.get_pdf(x, *par_array)
    scipy_y = scipy_crystal_ball.pdf(x, beta, m, mu, sigma)

    assert isinstance(crystal_ball, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    x_lo = -10000
    x_hi = 10000

    y_sig, y_ext = crystal_ball.pdf_ext(x, x_lo, x_hi, n_sig, *par_array)
    assert np.allclose(y_ext, n_sig * scipy_y, rtol=1e-8)
    assert np.allclose(y_sig, n_sig, rtol=1e-8)

    normalization = np.diff(
        scipy_crystal_ball.cdf(x[np.array([0, -1])], beta, m, mu, sigma)
    )
    y_norm = crystal_ball.pdf_norm(x, x[0], x[-1], *par_array)

    assert np.allclose(y_norm, scipy_y / normalization, rtol=1e-8)


def test_crystalball_cdf():
    x = np.arange(-10, 12)
    beta = 2
    m = 3
    mu = 1
    sigma = 2

    par_array = [mu, sigma, beta, m]

    y = crystal_ball.cdf(x, *par_array)
    y_direct = crystal_ball.get_cdf(x, *par_array)
    scipy_y = scipy_crystal_ball.cdf(x, beta, m, mu, sigma)

    assert isinstance(crystal_ball, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    y_ext = crystal_ball.cdf_ext(x, n_sig, *par_array)
    assert np.allclose(y_ext, 20 * scipy_y, rtol=1e-8)

    normalization = np.diff(
        scipy_crystal_ball.cdf(x[np.array([0, -1])], beta, m, mu, sigma)
    )
    y_norm = crystal_ball.cdf_norm(x, x[0], x[-1], *par_array)

    assert np.allclose(y_norm, scipy_y / normalization, rtol=1e-8)


def test_required_args():
    names = crystal_ball.required_args()
    assert names[0] == "mu"
    assert names[1] == "sigma"
    assert names[2] == "beta"
    assert names[3] == "m"


def test_name():
    assert crystal_ball.name == "crystal_ball"



from __future__ import annotations

import logging
from types import FunctionType

import numpy as np
import pandas as pd
from iminuit import Minuit, cost
from lgdo import lh5

log = logging.getLogger(__name__)


def convert_to_minuit(pars, func):
    try:
        c = cost.UnbinnedNLL(np.array([0]), func.pdf_ext)
    except AttributeError:
        c = cost.UnbinnedNLL(np.array([0]), func)
    if isinstance(pars, dict):
        m = Minuit(c, **pars)
    else:
        m = Minuit(c, *pars)
    return m


def return_nans(input):
    if isinstance(input, FunctionType):
        args = input.__code__.co_varnames[: input.__code__.co_argcount][1:]
        m = convert_to_minuit(np.full(len(args), np.nan), input)
        return m.values, m.errors, np.full((len(m.values), len(m.values)), np.nan)
    else:
        args = input.required_args()
        m = convert_to_minuit(np.full(len(args), np.nan), input)
        return m.values, m.errors, np.full((len(m.values), len(m.values)), np.nan)


def load_data(
    files: str | list | dict,
    lh5_path: str,
    cal_dict: dict,
    params: set,
    cal_energy_param: str = "cuspEmax_ctc_cal",
    threshold=None,
    return_selection_mask=False,
) -> pd.DataFrame | tuple(pd.DataFrame, np.array):
    """
    Loads parameters from data files. Applies calibration to cal_energy_param
    and uses this to apply a lower energy threshold.

    files
        file or list of files or dict pointing from timestamps to lists of files
    lh5_path
        path to table in files
    cal_dict
        dictionary with operations used to apply calibration constants
    params
        list of parameters to load from file
    cal_energy_param
        name of uncalibrated energy parameter
    threshold
        lower energy threshold for events to load
    return_selection_map
        if True, return selection mask for threshold along with data
    """

    params = set(params)
    if isinstance(files, str):
        files = [files]

    if isinstance(files, dict):
        # Go through each tstamp and recursively load_data on file lists
        df = []
        masks = []
        for tstamp, tfiles in files.items():
            file_df = load_data(
                tfiles,
                lh5_path,
                cal_dict.get(tstamp, cal_dict),
                params,
                cal_energy_param,
                threshold,
                return_selection_mask,
            )

            if return_selection_mask:
                file_df[0]["run_timestamp"] = np.full(
                    len(file_df[0]), tstamp, dtype=object
                )
                df.append(file_df[0])
                masks.append(file_df[1])
            else:
                file_df["run_timestamp"] = np.full(len(file_df), tstamp, dtype=object)
                df.append(file_df)

        df = pd.concat(df)
        if return_selection_mask:
            masks = np.concatenate(masks)

    elif isinstance(files, list):
        # Get set of available fields between input table and cal_dict
        file_keys = lh5.ls(
            files[0], lh5_path if lh5_path[-1] == "/" else lh5_path + "/"
        )
        file_keys = {key.split("/")[-1] for key in file_keys}

        # Get set of keys in calibration expressions that show up in file
        cal_keys = {
            name
            for info in cal_dict.values()
            for name in compile(info["expression"], "0vbb is real!", "eval").co_names
        } & file_keys

        # Get set of fields to read from files
        fields = cal_keys | (file_keys & params)

        lh5_it = lh5.iterator.LH5Iterator(
            files, lh5_path, field_mask=fields, buffer_len=100000
        )
        df_fields = params & (fields | set(cal_dict))
        if df_fields != params:
            log.debug(
                f"load_data(): params not found in data files or cal_dict: {params-df_fields}"
            )
        df = pd.DataFrame(columns=list(df_fields))

        for table in lh5_it:
            # Evaluate all provided expressions and add to table
            for outname, info in cal_dict.items():
                table[outname] = table.eval(
                    info["expression"], info.get("parameters", None)
                )
            entry = lh5_it.current_global_entries[0]
            n_rows = len(table)

            # Copy params in table into dataframe
            for par in df:
                # First set of entries: allocate enough memory for all entries
                if entry == 0:
                    df[par] = np.resize(table[par], len(lh5_it))
                else:
                    df.loc[entry : entry + n_rows - 1, par] = table[par][:n_rows]

        # Evaluate threshold mask and drop events below threshold
        if threshold is not None:
            masks = df[cal_energy_param] > threshold
            df.drop(np.where(~masks)[0], inplace=True)
        else:
            masks = np.ones(len(df), dtype=bool)

    log.debug("data loaded")
    if return_selection_mask:
        return df, masks
    else:
        return df



"""
This module is for extracting the pole zero constants from the decay tail
"""

from __future__ import annotations

import logging

import lgdo
import matplotlib.pyplot as plt
import numpy as np
from iminuit import Minuit
from scipy.stats import linregress

import pygama.pargen.dsp_optimize as opt
from pygama.pargen.data_cleaning import get_mode_stdev

log = logging.getLogger(__name__)


def dpz_model(
    ts: np.array, amp: float, tau1: float, tau2: float, f2: float
) -> np.array:
    """
    Models the double pole zero function as A*[(1-f2)*exp(-t/tau1) + f2*exp(-t/tau2)],
    this is the expected shape of an HPGe waveform. Used in performing fits in :func:`dpz_model_fit`

    Parameters
    ----------
    ts
        Array of time point values, in samples
    A
        The overall amplitude of the waveform
    tau1
        The short time constant present
    tau2
        The long time constant present
    f2
        The fraction of the long time constant present in the overall waveform
    """
    return amp * (1 - f2) * np.exp(-ts / tau1) + amp * f2 * np.exp(-ts / tau2)


def linear_dpz_fit(
    waveform: np.array,
    percent_tau1_fit: float,
    percent_tau2_fit: float,
    idx_shift: int,
    plot: int,
) -> tuple[float, float, float, dict]:
    """
    Parameters
    ----------
    waveform
        An array containing waveform data
    percent_tau1_fit
        The fractional percent of the length of the tail of the waveform to fit, used to fit the start of the tail.
    percent_tau2_fit
        The fractional percent of the length of the tail of the waveform to fit, used to fit the end of the tail.
    idx_shift
        The number of indices off the maximum of the waveform to start the fit of the tail. Usually ~100 samples works fine.
    plot
        An integer, if greater than 1 plots and shows the fit results, if greater than 0 saves the plot to a dictionary.

    Returns
    -------
    tau1
        A guess of the shorter double pole zero time constant
    tau2
        A guess of the longer double pole zero time constant
    frac
        A guess of the fraction in the double pole zero of the longer time constant
    out_plot_dict
        A dictionary containing the output plots if requested

    Notes
    -----
    Extracts an initial guess of the double pole zero parameters by performing two linear fits to the log of the waveform.
    The first log-fit is to long time scales, and fits the long time constant tau2. This piece is subtracted off the waveform,
    and then the shorter time constant is fit. The average of the fractions returned by the intercepts of these fits is returned.

    The waveform must be baseline subtracted to provide accurate fit results!
    """
    out_plot_dict = {}
    # Get the indices to start the fit from
    wf_max_idx = np.argmax(waveform)
    tau2_idx = -int((len(waveform) - wf_max_idx) * percent_tau2_fit)
    tau1_idx = int((len(waveform) - wf_max_idx) * percent_tau1_fit)

    # Rescale the waveform to ensure fitting converges
    fit_start_idx = (
        np.argmax(waveform) + idx_shift
    )  # shifting from the max of the waveform helps the convervence
    scaled_pulse = waveform / np.amax(waveform)
    scaled_pulse = scaled_pulse[fit_start_idx:]

    # Fit the long time constant
    ts = np.arange(0, len(scaled_pulse))
    slope_2, intercept_2, *_ = linregress(
        ts[tau2_idx:], np.log(scaled_pulse[tau2_idx:])
    )

    if plot > 0:
        fig = plt.figure(figsize=(12, 8))
        plt.plot(ts[tau2_idx:], np.log(scaled_pulse[tau2_idx:]))
        plt.plot(ts[tau2_idx:], slope_2 * ts[tau2_idx:] + intercept_2)
        plt.ylabel("ADC")
        plt.xlabel("Time [Samples]")
        plt.title("Long Time Constant Fitted")
        plt.grid(True)
        out_plot_dict["long_tau_linear"] = fig
        if plot > 1:
            plt.show()
        else:
            plt.close()

    # Subtract off the long time constant
    sub_pulse = np.abs(scaled_pulse - np.exp(intercept_2) * np.exp(ts * slope_2))

    # Fit the short time constant
    slope_1, intercept_1, *_ = linregress(ts[:tau1_idx], np.log(sub_pulse[:tau1_idx]))
    # If the fit fails, just return an arbitrary good-enough guess. All detectors have a roughly ~130 sample short time constant for presumming of 8.
    if slope_1 > 0:
        slope_1 = -1 / 130
        intercept_1 = np.log(1 - np.exp(intercept_2))

    if plot > 0:
        fig = plt.figure(figsize=(12, 8))
        plt.plot(ts[:tau1_idx], np.log(sub_pulse[:tau1_idx]))
        plt.plot(ts[:tau1_idx], slope_1 * ts[:tau1_idx] + intercept_1)
        plt.ylabel("ADC")
        plt.xlabel("Time [Samples]")
        plt.title("Short Time Constant Fitted")
        plt.grid(True)
        out_plot_dict["short_tau_linear"] = fig
        if plot > 1:
            plt.show()
        else:
            plt.close()

    return (
        -1 / slope_1,
        -1 / slope_2,
        np.mean([np.exp(intercept_2), 1 - np.exp(intercept_1)]),
        out_plot_dict,
    )


def dpz_model_fit(
    waveform: np.array,
    percent_tau1_fit: float,
    percent_tau2_fit: float,
    idx_shift: int,
    plot: int,
) -> tuple[float, float, float, dict]:
    """
    Parameters
    ----------
    waveform
        An array containing waveform data that has been baseline subtracted
    percent_tau1_fit
        The fractional percent of the length of the tail of the waveform to fit, used to fit the start of the tail.
    percent_tau2_fit
        The fractional percent of the length of the tail of the waveform to fit, used to fit the end of the tail.
    idx_shift
        The number of indices off the maximum of the waveform to start the fit of the tail. Usually ~100 samples works fine.
    plot
        An integer, that if greater than 1 plots and shows the fit results, if greater than 0 saves the plots to a dictionary

    Returns
    -------
    tau1
        A guess of the shorter double pole zero time constant
    tau2
        A guess of the longer double pole zero time constant
    frac2
        A guess of the fraction in the double pole zero of the longer time constant
    out_plot_dict
        A dictionary containing the output plots if requested

    Notes
    -----
    Extracts the double pole zero parameters by fitting to an analytic model, utilizing the best-guess from linear fitting.
    Fits to the function
    f(t) = A*[(1-f2)*exp(-t/tau1)+f2*exp(-t/tau2)]
    It fits from the maximum of the waveform onwards to the end of the tail.
    The waveform must be baseline subtracted in order to get the best fit results
    """
    out_plot_dict = {}
    # Get the initial guess of the DPZ constants using a quick linear fit to the start and end of the tail
    tau1_guess, tau2_guess, f2_guess, linear_plot_dict = linear_dpz_fit(
        waveform, percent_tau1_fit, percent_tau2_fit, idx_shift, plot
    )
    out_plot_dict |= linear_plot_dict  # merge dictionary in place

    # Select the data to fit, from the maximum plus some offset
    fit_start_idx = np.argmax(waveform) + idx_shift
    waveform = waveform[fit_start_idx:]

    # Create the Iminuit cost function using least squares
    ts = np.arange(0, len(waveform))

    def cost_function(amp, tau1, tau2, f2):
        output = dpz_model(ts, amp, tau1, tau2, f2)
        res = 0
        for i in range(len(output)):
            res += (output[i] - waveform[i]) ** 2

        return res

    # Perform the fit
    m = Minuit(
        cost_function,
        amp=np.amax(waveform),
        tau1=tau1_guess,
        tau2=tau2_guess,
        f2=f2_guess,
    )
    m.errordef = Minuit.LEAST_SQUARES
    m.limits[0] = (0, None)
    m.limits[1] = (0, None)
    m.limits[2] = (0, None)
    m.limits[3] = (0, 1)  # the fraction MUST be between 0 and 1
    m.migrad()

    if plot > 0:
        fig = plt.figure(figsize=(12, 8))
        plt.scatter(ts, waveform)
        plt.plot(ts, dpz_model(ts, *m.values), c="r")
        plt.ylabel("ADC")
        plt.xlabel("Time [Samples]")
        plt.title("DPZ Model Fit to Waveform")
        plt.grid(True)
        out_plot_dict["dpz_model_fit"] = fig
    if plot > 1:
        plt.show()
    else:
        plt.close()

    if plot > 0:
        fig = plt.figure(figsize=(12, 8))
        plt.scatter(ts, waveform - dpz_model(ts, *m.values), c="r")
        plt.ylabel("ADC")
        plt.xlabel("Time [Samples]")
        plt.title("DPZ Model Waveform Fit Residuals")
        plt.grid(True)
        out_plot_dict["dpz_model_fit_residuals"] = fig
    if plot > 1:
        plt.show()
    else:
        plt.close()

    # order so tau1 largest
    if m.values[1] > m.values[2]:
        return m.values[1], m.values[2], m.values[3], out_plot_dict
    else:
        return m.values[2], m.values[1], 1 - m.values[3], out_plot_dict


def tp100_align(wfs: np.array, tp100_window_width: int, tp100s: np.array) -> np.array:
    """
    Align provided waveforms at their maximum, so that an average over all the waveforms creates a valid superpulse to fit using :func:`dpz_model_fit`.
    Do this without sacrificing too much of the length of the decaying tail

    Parameters
    ----------
    wfs
        An array of arrays containing waveforms
    tp100_window_width
        The window of acceptance, with a center set at the median tp100 value. If a tp100 falls outside this window, ignore this waveform for the superpulse
    tp100s
        An array containing the indices of the maximums of the waveforms, in samples

    Returns
    -------
    time_aligned_wfs
        An array of waveforms that are all aligned at their maximal values
    """
    tp100_window_width = (
        13  # If tp100 isn't in +/- this window of the median, forget about it
    )
    median_tp100 = int(np.nanmedian(tp100s))  # in samples
    wf_len = len(wfs[0])
    time_aligned_wfs = []

    for i, wf in enumerate(wfs):
        if np.isnan(tp100s[i]):
            pass
        elif (
            median_tp100 - tp100_window_width
            <= tp100s[i]
            <= median_tp100 + tp100_window_width
        ):
            wf_win = wf[
                tp100s[i]
                - (median_tp100 - tp100_window_width) : wf_len
                - (-1 * tp100s[i] + median_tp100 + tp100_window_width)
            ]
            time_aligned_wfs.append(wf_win)

    return time_aligned_wfs


class PZCorrect:
    def __init__(self, dsp_config, wf_field, debug_mode=False):
        self.dsp_config = dsp_config
        self.wf_field = wf_field
        self.output_dict = {}
        self.results_dict = {}
        self.debug_mode = debug_mode

    def get_single_decay_constant(
        self, tb_data: lgdo.Table, slope_param="tail_slope", display=0
    ):
        """
        Finds the decay constant from the modal value of the tail slope after cuts
        and saves it to the specified json. Updates self.output_dict with tau value

        Parameters
        ----------
        - slopes: numpy array of tail slopes
        - wfs: WaveformTable object containing waveform data

        """
        tb_out = opt.run_one_dsp(tb_data, self.dsp_config)
        slopes = tb_out[slope_param].nda
        wfs = tb_data[self.wf_field]

        mode, stdev = get_mode_stdev(slopes)
        tau = round(-1 / (mode), 1)
        err = round((-1 / (mode + (stdev / np.sqrt(len(slopes))))) - tau, 1)

        sampling_rate = wfs["dt"].nda[0]
        units = wfs["dt"].attrs["units"]
        tau = f"{tau*sampling_rate}*{units}"
        err = f"{err*sampling_rate}*{units}"

        if "pz" in self.output_dict:
            self.output_dict["pz"].update({"tau1": tau, "tau1_err": err})
        else:
            self.output_dict["pz"] = {"tau1": tau, "tau1_err": err}

        self.results_dict.update(
            {"single_decay_constant": {"slope_pars": {"mode": mode, "stdev": stdev}}}
        )

        if display <= 0:
            return
        else:
            out_plot_dict = {}
            return out_plot_dict

    def get_dpz_decay_constants(
        self,
        tb_data: lgdo.Table,
        percent_tau1_fit: float = 0.1,
        percent_tau2_fit: float = 0.2,
        offset_from_wf_max: int = 10,
        superpulse_bl_idx: int = 25,
        superpulse_window_width: int = 13,
        display: int = 0,
    ) -> dict:
        """
        Gets values for the DPZ time constants in 3 stages:

        1. Perform a linear fit to the start and end of the decaying tail of a superpulse
        2. Use those initial guesses to seed a LSQ fit to a DPZ model of the sum of two decaying exponentials
        3. Use the results of the model fit as initial guesses in a DSP routine that optimizes the flatness of the decaying tail

        The first step in this process is to generate a superpulse from high-energy waveforms.


        Parameters
        ----------
        tb_data
            Table containing high-energy event waveforms and daqenergy. Pulser events must have already been removed from this table
        percent_tau1_fit
            The fractional percent of the length of the tail of the waveform to fit, used to fit the start of the tail.
        percent_tau2_fit
            The fractional percent of the length of the tail of the waveform to fit, used to fit the end of the tail.
        offset_from_wf_max
            The number of indices off the maximum of the waveform to start the fit of the tail. Usually ~100 samples works fine.
        superpulse_bl_idx
            The index at which to stop when computing the mean value of a baseline, used for DPZ
        superpulse_window_width
            The window of acceptance for tp100s while selecting waveforms to make a superpulse, with a center set at the median tp100 value. If a tp100 falls outside this window, ignore this waveform for the superpulse
        display
            An integer. If greater than 1, plots and shows the attempts to fit the long and short time constants. If greater than 0, saves the plot to a dictionary.


        Returns
        -------
        tau_dict
            dictionary of form {'dpz': {'tau1': decay_constant1, 'tau2': decay_constant2, 'frac': fraction}}
        out_plot_dict
            A dictionary containing monitoring plots


        Notes
        -----
        tau1 is the shorter time constant, tau2 is the longer, and frac is the amount of the larger time constant present in the sum of the two exponentials
        """

        # Get high energy waveforms to create a superpulse. Eventually allow user which peak to select? For now, use the 2615 keV peak
        sampling_rate = tb_data[self.wf_field]["dt"].nda[0]
        units = tb_data[self.wf_field]["dt"].attrs["units"]

        high_e_wfs = tb_data[self.wf_field]["values"].nda[:]

        # Time align the waveforms to their maximum
        tp100s = []
        for wf in high_e_wfs:
            tp100s.append(np.argmax(wf))

        time_aligned_wfs = tp100_align(high_e_wfs, superpulse_window_width, tp100s)

        # Baseline subtract the time aligned waveforms
        bl_sub_time_aligned_wfs = []

        for i in range(len(time_aligned_wfs)):
            bl_sub_time_aligned_wfs.append(
                time_aligned_wfs[i] - np.mean(time_aligned_wfs[i][:superpulse_bl_idx])
            )

        # Create a superpulse
        superpulse = np.mean(bl_sub_time_aligned_wfs, axis=0)

        # Fit the superpulse and get rough DPZ constants
        tau1s_fit, tau2s_fit, f2s_fit, out_plot_dict = dpz_model_fit(
            superpulse,
            percent_tau1_fit=percent_tau1_fit,
            percent_tau2_fit=percent_tau2_fit,
            idx_shift=offset_from_wf_max,
            plot=display,
        )
        log.debug("Found initial guesses for DPZ constants:")
        for item, value in {
            "tau1": f"{tau1s_fit * sampling_rate}*{units}",
            "tau2": f"{tau2s_fit * sampling_rate}*{units}",
            "frac": f2s_fit,
        }.items():
            log.debug(f"{item}: {value}")

        self.results_dict.update(
            {
                "double_pole_zero": {
                    "guesses": {"tau1": tau1s_fit, "tau2": tau2s_fit, "frac": f2s_fit}
                }
            }
        )

        # Optimize the flatness of high energy waveforms to get optimal DPZ constants
        dpz_opt_tb_out = opt.run_one_dsp(
            tb_data,
            self.dsp_config,
            db_dict=dict(
                {"pz": {"tau1": tau1s_fit, "tau2": tau2s_fit, "frac": f2s_fit}}
            ),
        )

        # Update tau_dict with the dpz constants
        tau1 = np.nanmedian(dpz_opt_tb_out["tau1"].nda)
        tau2 = np.nanmedian(dpz_opt_tb_out["tau2"].nda)
        frac = np.nanmedian(dpz_opt_tb_out["frac"].nda)
        tau1_err = (
            np.nanpercentile(dpz_opt_tb_out["tau1"].nda, 68.27)
            - np.nanpercentile(dpz_opt_tb_out["tau1"].nda, 31.73)
        ) / 2
        tau2_err = (
            np.nanpercentile(dpz_opt_tb_out["tau2"].nda, 68.27)
            - np.nanpercentile(dpz_opt_tb_out["tau2"].nda, 31.73)
        ) / 2
        frac_err = (
            np.nanpercentile(dpz_opt_tb_out["frac"].nda, 68.27)
            - np.nanpercentile(dpz_opt_tb_out["frac"].nda, 31.73)
        ) / 2

        if "units" in dpz_opt_tb_out["tau1"].attrs and dpz_opt_tb_out["tau1"].attrs[
            "units"
        ] not in ["ADC", "sample"]:
            tau1 = f'{tau1}*{dpz_opt_tb_out["tau1"].attrs["units"]}'
            tau1_err = f'{tau1_err}*{dpz_opt_tb_out["tau1"].attrs["units"]}'
        else:
            tau1 = f"{tau1*sampling_rate}*{units}"
            tau1_err = f"{tau1_err*sampling_rate}*{units}"

        if "units" in dpz_opt_tb_out["tau2"].attrs and dpz_opt_tb_out["tau2"].attrs[
            "units"
        ] not in ["ADC", "sample"]:
            tau2 = f'{tau2}*{dpz_opt_tb_out["tau2"].attrs["units"]}'
            tau2_err = f'{tau2_err}*{dpz_opt_tb_out["tau2"].attrs["units"]}'
        else:
            tau2 = f"{tau2*sampling_rate}*{units}"
            tau2_err = f"{tau2_err*sampling_rate}*{units}"

        output_dict = {
            "tau1": tau1,
            "tau2": tau2,
            "frac": frac,
            "tau1_err": tau1_err,
            "tau2_err": tau2_err,
            "frac_err": frac_err,
        }
        if "pz" in self.output_dict:
            self.output_dict["pz"].update(output_dict)
        else:
            self.output_dict["pz"] = output_dict

        if display <= 0:
            return
        else:
            return out_plot_dict

    def plot_waveforms_after_correction(
        self,
        tb_data,
        wf_field,
        xlim=(0, 1024),
        ylim=None,
        norm_param=None,
        display=0,
        figsize=(8, 6),
        fontsize=8,
    ):
        tb_out = opt.run_one_dsp(tb_data, self.dsp_config, db_dict=self.output_dict)
        wfs = tb_out[wf_field]["values"].nda
        wf_idxs = np.random.choice(len(wfs), 100)
        if norm_param is not None:
            means = tb_out[norm_param].nda[wf_idxs]
            wfs = np.divide(wfs[wf_idxs], np.reshape(means, (len(wf_idxs), 1)))
        else:
            wfs = wfs[wf_idxs]
        plt.rcParams["font.size"] = fontsize
        fig = plt.figure(figsize=figsize)
        for wf in wfs:
            plt.plot(np.arange(0, len(wf), 1), wf)
        plt.axhline(1, color="black")
        plt.axhline(0, color="black")
        plt.xlim(xlim)
        if ylim is not None:
            plt.ylim(ylim)
        plt.xlabel("Samples")
        plt.ylabel("ADU")
        plot_dict = {"waveforms": fig}
        if display > 1:
            plt.show()
        else:
            plt.close()
        return plot_dict

    def plot_slopes(
        self,
        tb_data,
        slope_field,
        with_correction=False,
        display=0,
        figsize=(8, 6),
        fontsize=8,
    ):

        tb_out = opt.run_one_dsp(
            tb_data,
            self.dsp_config,
            db_dict=self.output_dict if with_correction else {},
        )
        slopes = tb_out[slope_field].nda

        plt.rcParams["font.size"] = fontsize
        fig, ax = plt.subplots()
        fig.set_figheight(figsize[1])
        fig.set_figwidth(figsize[0])
        bins = np.arange(
            np.nanpercentile(slopes, 1),
            np.nanpercentile(slopes, 99),
            np.nanpercentile(slopes, 51) - np.nanpercentile(slopes, 50),
        )
        counts, bins, bars = ax.hist(slopes, bins=bins, histtype="step")
        plt.xlabel("Slope")
        plt.ylabel("Counts")
        if "single_decay_constant" in self.results_dict:
            high_bin = self.results_dict["single_decay_constant"]["slope_pars"]["mode"]
            sigma = self.results_dict["single_decay_constant"]["slope_pars"]["stdev"]
            ax.axvline(high_bin, color="red")
            in_min = high_bin - 4 * sigma
            in_max = high_bin + 4 * sigma
            axins = ax.inset_axes([0.6, 0.6, 0.4, 0.4])
            axins.hist(
                slopes[(slopes > in_min) & (slopes < in_max)],
                bins=50,
                histtype="step",
            )
            axins.axvline(high_bin, color="red")
            axins.set_xlim(in_min, in_max)
            ax.set_xlim(np.nanpercentile(slopes, 1), np.nanpercentile(slopes, 99))
        if with_correction:
            out_plot_dict = {"corrected_slope": fig}
        else:
            out_plot_dict = {"slope": fig}
        if display > 1:
            plt.show()
        else:
            plt.close()
        return out_plot_dict



import numpy as np
from scipy.stats import norm

from pygama.math.functions.pygama_continuous import PygamaContinuous
from pygama.math.functions.step import step


def test_step_pdf():
    x = np.arange(-10, 10)
    mu = 2.3
    sigma = 1.1
    hstep = 0.65
    x_lo = np.amin(x)
    x_hi = np.amax(x)

    param_array = [x_lo, x_hi, mu, sigma, hstep]

    y = step.pdf(x, *param_array)
    y_direct = step.get_pdf(x, *param_array)

    # compute the unnormalized step function
    scipy_step = 1 + hstep * (
        norm.cdf(x, mu, sigma) * 2 - 1
    )  # pdf = (1+erf(x/np.sqrt(2))); erf(x) = 2 cdf_norm(x*sqrt(2))-1

    # compute the normalization for the step function
    maximum = (x_hi - mu) / sigma
    minimum = (x_lo - mu) / sigma
    normalization = sigma * (
        (1 - hstep) * (maximum - minimum)
        + 2
        * hstep
        * (
            maximum * norm.cdf(maximum)
            + np.exp(-1 * maximum**2 / 2) / np.sqrt(2 * np.pi)
            - minimum * norm.cdf(minimum)
            - np.exp(-1 * minimum**2 / 2) / np.sqrt(2 * np.pi)
        )
    )

    # compute the normalized step function
    scipy_y = scipy_step / normalization

    assert isinstance(step, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    x_lo = -10
    x_hi = 9

    y_sig, y_ext = step.pdf_ext(x, x_lo, x_hi, n_sig, mu, sigma, hstep)
    assert np.allclose(y_ext, n_sig * scipy_y, rtol=1e-8)
    assert np.allclose(y_sig, n_sig, rtol=1e-8)

    y_norm = step.pdf_norm(x, x_lo, x_hi, mu, sigma, hstep)

    assert np.allclose(y_norm, scipy_y, rtol=1e-8)


def test_step_cdf():
    x = np.arange(-10, 10)
    mu = 1.1
    sigma = 2.3
    hstep = 4.5
    x_lo = np.amin(x)
    x_hi = np.amax(x)

    param_array = [x_lo, x_hi, mu, sigma, hstep]

    y = step.cdf(x, *param_array)
    y_direct = step.get_cdf(x, *param_array)

    # Compute the normalization of the pdf
    maximum = (np.amax(x) - mu) / sigma
    minimum = (np.amin(x) - mu) / sigma
    pdf_normalization = sigma * (
        (1 - hstep) * (maximum - minimum)
        + 2
        * hstep
        * (
            maximum * norm.cdf(maximum)
            + np.exp(-1 * maximum**2 / 2) / np.sqrt(2 * np.pi)
            - minimum * norm.cdf(minimum)
            - np.exp(-1 * minimum**2 / 2) / np.sqrt(2 * np.pi)
        )
    )

    z = (x - mu) / sigma
    # Compute the unnormalized cdf
    unnormalized_cdf = sigma * (
        (1 - hstep) * (z - minimum)
        + 2
        * hstep
        * (
            z * norm.cdf(z)
            + np.exp(-1 * z**2 / 2) / np.sqrt(2 * np.pi)
            - minimum * norm.cdf(minimum)
            - np.exp(-1 * minimum**2 / 2) / np.sqrt(2 * np.pi)
        )
    )

    # Compute the cdf
    scipy_y = unnormalized_cdf / pdf_normalization

    assert isinstance(step, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    y_ext = step.cdf_ext(x, x_lo, x_hi, n_sig, mu, sigma, hstep)
    assert np.allclose(y_ext, n_sig * scipy_y, rtol=1e-8)

    y_norm = step.cdf_norm(x, x[0], x[-1], mu, sigma, hstep)

    assert np.allclose(y_norm, scipy_y, rtol=1e-8)


def test_required_args():
    names = step.required_args()
    assert names[0] == "x_lo"
    assert names[1] == "x_hi"
    assert names[2] == "mu"
    assert names[3] == "sigma"
    assert names[4] == "hstep"


def test_name():
    assert step.name == "step"



import numpy as np
from scipy.stats import exponnorm

from pygama.math.functions.exgauss import exgauss
from pygama.math.functions.pygama_continuous import PygamaContinuous


def test_exgauss_pdf():
    x = np.arange(-10, 10)
    mu = 0.5
    sigma = 2
    tau = 2

    par_array = [mu, sigma, tau]

    y = exgauss.pdf(x, sigma, tau, mu, sigma)
    y_direct = exgauss.get_pdf(x, *par_array)
    scipy_y = exponnorm.pdf(
        -1 * x, tau / sigma, -1 * mu, sigma
    )  # to be equivalent to the scipy version, x -> -x, mu -> -mu, k -> k/sigma

    assert isinstance(exgauss, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 10
    x_lo = -100
    x_hi = 100

    y_sig, y_ext = exgauss.pdf_ext(x, x_lo, x_hi, n_sig, *par_array)
    assert np.allclose(y_ext, n_sig * scipy_y, rtol=1e-8)
    assert np.allclose(y_sig, n_sig, rtol=1e-8)

    normalization = np.diff(
        1 - exponnorm.cdf(-1 * x[np.array([0, -1])], tau / sigma, -1 * mu, sigma)
    )
    y_norm = exgauss.pdf_norm(x, x[0], x[-1], *par_array)

    assert np.allclose(y_norm, scipy_y / normalization, rtol=1e-8)


def test_exgauss_cdf():
    x = np.arange(-10, 10)
    mu = 0.5
    sigma = 2
    tau = 2

    par_array = [mu, sigma, tau]

    y = exgauss.cdf(x, sigma, tau, mu, sigma)
    scipy_y = 1 - exponnorm.cdf(
        -1 * x, tau / sigma, -1 * mu, sigma
    )  # to be equivalent to the scipy version, x -> -x, mu -> -mu, k -> k/sigma
    y_direct = exgauss.get_cdf(x, *par_array)

    assert isinstance(exgauss, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    y_ext = exgauss.cdf_ext(x, n_sig, *par_array)
    assert np.allclose(y_ext, n_sig * scipy_y, rtol=1e-8)

    normalization = np.diff(scipy_y[np.array([0, -1])])
    y_norm = exgauss.cdf_norm(x, x[0], x[-1], *par_array)
    assert np.allclose(y_norm, scipy_y / normalization, rtol=1e-8)


def test_required_args():
    names = exgauss.required_args()
    assert names[0] == "mu"
    assert names[1] == "sigma"
    assert names[2] == "tau"


def test_name():
    assert exgauss.name == "exgauss"



import numpy as np
import pytest

from pygama.evt.modules import larveto


def test_tc_time_pdf():
    assert isinstance(larveto.l200_tc_time_pdf(0), float)
    assert isinstance(
        larveto.l200_tc_time_pdf(np.array([0, -0.5, 3]) * 1e3), np.ndarray
    )
    assert len(larveto.l200_tc_time_pdf(np.array([]))) == 0

    with pytest.raises(ValueError):
        larveto.l200_tc_time_pdf([-10_000, 42])
    with pytest.raises(ValueError):
        larveto.l200_tc_time_pdf([0, 50_000])



from __future__ import annotations

import json
import logging
import os
from collections.abc import MutableMapping
from pathlib import Path
from typing import Any, Iterator

import yaml

log = logging.getLogger(__name__)


def getenv_bool(name: str, default: bool = False) -> bool:
    """Get environment value as a boolean, returning True for 1, t and true
    (caps-insensitive), and False for any other value and default if undefined.
    """
    val = os.getenv(name)
    if not val:
        return default
    elif val.lower() in ("1", "t", "true"):
        return True
    else:
        return False


class NumbaPygamaDefaults(MutableMapping):
    """Bare-bones class to store some Numba default options. Defaults values
    are set from environment variables. Useful for the pygama.math distributions

    Examples
    --------
    Set all default option values for a numba wrapped function at once by expanding the
    provided dictionary:

    >>> from numba import njit
    >>> from pygama.utils import numba_math_defaults_kwargs as nb_kwargs
    >>> @njit([], "", **nb_kwargs, nopython=True) # def dist(...): ...

    Customize one argument but still set defaults for the others:

    >>> from pygama.utils import numba_math_defaults as nb_defaults
    >>> @njit([], "", **nb_defaults(cache=False) # def dist(...): ...

    Override global options at runtime:

    >>> from pygama.utils import numba_math_defaults
    >>> # must set options before explicitly importing pygama.math.distributions!
    >>> numba_math_defaults.cache = False
    """

    def __init__(self) -> None:
        self.parallel: bool = getenv_bool("PYGAMA_PARALLEL", default=False)
        self.fastmath: bool = getenv_bool("PYGAMA_FASTMATH", default=True)

    def __getitem__(self, item: str) -> Any:
        return self.__dict__[item]

    def __setitem__(self, item: str, val: Any) -> None:
        self.__dict__[item] = val

    def __delitem__(self, item: str) -> None:
        del self.__dict__[item]

    def __iter__(self) -> Iterator:
        return self.__dict__.__iter__()

    def __len__(self) -> int:
        return len(self.__dict__)

    def __call__(self, **kwargs) -> dict:
        mapping = self.__dict__.copy()
        mapping.update(**kwargs)
        return mapping

    def __str__(self) -> str:
        return str(self.__dict__)

    def __repr__(self) -> str:
        return str(self.__dict__)


numba_math_defaults = NumbaPygamaDefaults()
numba_math_defaults_kwargs = numba_math_defaults

__file_extensions__ = {"json": [".json"], "yaml": [".yaml", ".yml"]}


def load_dict(fname: str, ftype: str | None = None) -> dict:
    """Load a text file as a Python dict."""
    fname = Path(fname)

    # determine file type from extension
    if ftype is None:
        for _ftype, exts in __file_extensions__.items():
            if fname.suffix in exts:
                ftype = _ftype

    msg = f"loading {ftype} dict from: {fname}"
    log.debug(msg)

    with fname.open() as f:
        if ftype == "json":
            return json.load(f)
        if ftype == "yaml":
            return yaml.safe_load(f)

        msg = f"unsupported file format {ftype}"
        raise NotImplementedError(msg)



r"""
Provide a convenience function for three gaussians on two steps

Parameters
----------
x_lo
    The lower range to compute the normalization of the step functions
x_hi
    The upper range to compute the normalization of the step functions
area1, mu_1, sigma_1
    The area, location, and scale of the first Gaussian
area2, mu_2, sigma_2
    The area, location, and scale of the second Gaussian
area3, mu_3, sigma_3
    The area, location, and scale of the third Gaussian
area4, hstep_1
    The area and height of the first step function
area5, hstep_2
    The area and height of the second step function

Example
-------
triple_gauss_on_double_step.get_pdf(x, pars = [x_lo, x_hi, n_sig1, mu1, sigma1, n_sig2, mu2, sigma2, n_sig3, mu3, sigma3, n_bkg1, hstep1, n_bkg2, hstep2])

Returns
-------
triple_gauss_on_double_step
    A subclass of SumDists and rv_continuous, has methods of pdf, cdf, etc.

Notes
-----
The first step function shares the mu_1, sigma_1 with the first Gaussian,
and the second step function shares the mu_2, sigma_2 with the second Gaussian
"""

from pygama.math.functions.gauss import gaussian
from pygama.math.functions.gauss_on_step import gauss_on_step
from pygama.math.functions.sum_dists import SumDists

(x_lo, x_hi, area1, mu1, sigma1, area4, hstep1, area2, mu2, sigma2, area5, hstep2) = (
    range(12)
)
par_array = [
    (gauss_on_step, [x_lo, x_hi, area1, mu1, sigma1, area4, hstep1]),
    (gauss_on_step, [x_lo, x_hi, area2, mu2, sigma2, area5, hstep2]),
]

double_gauss_on_double_step = SumDists(
    par_array,
    [],
    None,
    parameter_names=[
        "x_lo",
        "x_hi",
        "area1",
        "mu1",
        "sigma1",
        "area4",
        "hstep1",
        "area2",
        "mu2",
        "sigma2",
        "area5",
        "hstep2",
    ],
)

(
    x_lo,
    x_hi,
    area1,
    mu1,
    sigma1,
    area2,
    mu2,
    sigma2,
    area3,
    mu3,
    sigma3,
    area4,
    hstep1,
    area5,
    hstep2,
) = range(15)

par_array = [
    (gaussian, [mu3, sigma3]),
    (
        double_gauss_on_double_step,
        [
            x_lo,
            x_hi,
            area1,
            mu1,
            sigma1,
            area4,
            hstep1,
            area2,
            mu2,
            sigma2,
            area5,
            hstep2,
        ],
    ),
]
triple_gauss_on_double_step = SumDists(
    par_array,
    [area3],
    "one_area",
    parameter_names=[
        "x_lo",
        "x_hi",
        "n_sig1",
        "mu1",
        "sigma1",
        "n_sig2",
        "mu2",
        "sigma2",
        "n_sig3",
        "mu3",
        "sigma3",
        "n_bkg1",
        "hstep1",
        "n_bkg2",
        "hstep2",
    ],
    name="triple_gauss_on_double_step",
)



import lgdo.lh5 as lh5
import numpy as np
from pytest import approx

from pygama.math.distributions import hpge_peak
from pygama.pargen import energy_cal


def test_peak_match():
    expected_peaks_kev = [1460, 2614.5]
    peaks_adu = [
        78.676315,
        288.4798,
        603.18506,
        1337.4973,
        2019.3586,
        3225.7288,
        3907.59,
        5795.8213,
        8470.816,
    ]
    pars, best_ixtup, best_iytup = energy_cal.poly_match(
        peaks_adu, expected_peaks_kev, deg=0, atol=10
    )
    assert np.array_equal(best_ixtup, [5, 7])


def test_hpge_cal(lgnd_test_data):
    # test the HPGe calibration function
    # the function should return a calibration polynomial
    # that maps ADC channel to energy in keV

    # load test data here
    data = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/dsp/cal/p03/r000/l200-p03-r000-cal-20230311T235840Z-tier_dsp.lh5"
    )

    energy = lh5.read_as("ch1104000/dsp/cuspEmax", data, "np")

    glines = [860.564, 1592.53, 1620.50, 2103.53, 2614.50]

    pk_pars = [
        (860.564, (20, 20), hpge_peak),
        (1592.53, (20, 20), hpge_peak),
        (1620.50, (20, 20), hpge_peak),
        (2103.53, (20, 20), hpge_peak),
        (2614.50, (20, 20), hpge_peak),
    ]

    # test init
    cal = energy_cal.HPGeCalibration(
        "cuspEmax",
        glines,
        2615 / np.nanpercentile(energy, 99),
        deg=0,
        debug_mode=True,
    )

    # test dictionary generation
    out_dict = cal.gen_pars_dict()
    assert out_dict == {
        "expression": "a + b * cuspEmax",
        "parameters": {"a": 0.0, "b": 2615 / np.nanpercentile(energy, 99)},
    }

    cal.hpge_find_energy_peaks(energy, update_cal_pars=False)

    assert (cal.peaks_kev == glines).all()
    assert approx(cal.pars[1], 1) == 0.15
    assert cal.pars[0] == 0.0

    cal.hpge_find_energy_peaks(energy)

    assert len(cal.peaks_kev) == len(glines) and (cal.peaks_kev == glines).all()
    assert approx(cal.pars[1], 0.1) == 0.15
    assert cal.pars[0] == 0.0

    cal.hpge_get_energy_peaks(energy)

    assert len(cal.peaks_kev) == len(glines) and (cal.peaks_kev == glines).all()
    assert approx(cal.pars[1], 0.1) == 0.15
    assert cal.pars[0] == 0.0
    locs = cal.peak_locs.copy()
    cal.hpge_cal_energy_peak_tops(energy)

    assert len(cal.peaks_kev) == len(glines) and (cal.peaks_kev == glines).all()
    assert approx(cal.pars[1], 0.1) == 0.15
    assert cal.pars[0] == 0.0

    cal.peak_locs = locs
    cal.hpge_fit_energy_peaks(energy, peak_pars=pk_pars)

    assert len(cal.peaks_kev) == len(glines) and (cal.peaks_kev == glines).all()
    assert approx(cal.pars[1], 0.1) == 0.15
    assert cal.pars[0] == 0.0

    cal.get_energy_res_curve(
        energy_cal.FWHMLinear,
        interp_energy_kev={"Qbb": 2039.0},
    )

    assert (
        approx(
            cal.results["hpge_fit_energy_peaks"]["FWHMLinear"]["Qbb_fwhm_in_kev"], 0.1
        )
        == 2.3
    )


def test_hpge_cal_full_calibration(lgnd_test_data):
    data = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/dsp/cal/p03/r000/l200-p03-r000-cal-20230311T235840Z-tier_dsp.lh5"
    )

    energy = lh5.read_as("ch1104000/dsp/cuspEmax", data, "np")

    glines = [860.564, 1592.53, 1620.50, 2103.53, 2614.50]

    pk_pars = [
        (860.564, (20, 20), hpge_peak),
        (1592.53, (20, 20), hpge_peak),
        (1620.50, (20, 20), hpge_peak),
        (2103.53, (20, 20), hpge_peak),
        (2614.50, (20, 20), hpge_peak),
    ]

    cal = energy_cal.HPGeCalibration(
        "cuspEmax",
        glines,
        2615 / np.nanpercentile(energy, 99),
        deg=0,
        debug_mode=True,
    )

    cal.full_calibration(energy, peak_pars=pk_pars)

    assert len(cal.peaks_kev) == len(glines) and (cal.peaks_kev == glines).all()
    assert approx(cal.pars[1], 0.1) == 0.15
    assert cal.pars[0] == 0.0


def test_hpge_cal_prominent_peak(lgnd_test_data):
    data = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/dsp/cal/p03/r000/l200-p03-r000-cal-20230311T235840Z-tier_dsp.lh5"
    )

    energy = lh5.read_as("ch1104000/dsp/cuspEmax", data, "np")

    glines = [860.564, 1592.53, 1620.50, 2103.53, 2614.50]

    pk_pars = [
        (860.564, (20, 20), hpge_peak),
        (1592.53, (20, 20), hpge_peak),
        (1620.50, (20, 20), hpge_peak),
        (2103.53, (20, 20), hpge_peak),
        (2614.50, (20, 20), hpge_peak),
    ]

    # test in
    cal = energy_cal.HPGeCalibration(
        "cuspEmax",
        glines,
        2615 / np.nanpercentile(energy, 99),
        deg=0,
        debug_mode=True,
    )

    cal.calibrate_prominent_peak(energy, 2614.5, pk_pars)
    assert cal.peaks_kev[0] == 2614.5 and len(cal.peaks_kev) == 1
    assert approx(cal.pars[1], 0.1) == 0.15



import lgdo.lh5 as lh5
from pytest import approx

import pygama.pargen.data_cleaning as dc


def test_cuts(lgnd_test_data):
    # load test data here
    data = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/dsp/cal/p03/r000/l200-p03-r000-cal-20230311T235840Z-tier_dsp.lh5"
    )
    tbl = lh5.read("ch1104000/dsp", data)

    cut_pars = {
        "bl_std_cut": {"cut_parameter": "bl_std", "cut_level": 4, "mode": "inclusive"},
        "dt_eff_cut": {
            "expression": "(dt_eff>a)&(dt_eff<b)",
            "parameters": {"a": 0, "b": 4000},
        },
        "bl_pileup_cut": {
            "cut_parameter": "baselineEmax",
            "cut_level": 4,
            "mode": "exclusive",
        },
    }

    cut_dict = dc.generate_cuts(tbl, cut_pars)

    assert cut_dict["dt_eff_cut"] == cut_pars["dt_eff_cut"]
    assert cut_dict["bl_std_cut"]["expression"] == "(bl_std>a) & (bl_std<b)"
    assert approx(cut_dict["bl_std_cut"]["parameters"]["a"], 0.1) == 10
    assert approx(cut_dict["bl_std_cut"]["parameters"]["b"], 0.1) == 15
    assert (
        cut_dict["bl_pileup_cut"]["expression"] == "(baselineEmax<a) | (baselineEmax>b)"
    )
    assert approx(cut_dict["bl_pileup_cut"]["parameters"]["a"], 0.1) == 7
    assert approx(cut_dict["bl_pileup_cut"]["parameters"]["b"], 0.1) == 24

    # check also works for df
    df = lh5.read_as("ch1104000/dsp", data, "pd")

    df["baselineEmax"].to_numpy()

    cut_dict_df = dc.generate_cuts(df, cut_pars)

    assert cut_dict_df == cut_dict

    ids = dc.get_cut_indexes(df, cut_pars)
    ids_tbl = dc.get_cut_indexes(tbl, cut_pars)

    assert (ids == ids_tbl).all()



import pygama.utils as pgu


def test_math_numba_defaults():
    assert pgu.numba_math_defaults_kwargs.fastmath
    assert not pgu.numba_math_defaults_kwargs.parallel

    pgu.numba_math_defaults.fastmath = False
    assert not pgu.numba_math_defaults.fastmath



"""
pygama convenience functions for fitting ubinned data
"""

import logging
from typing import Callable

import numpy as np
from iminuit import Minuit, cost

log = logging.getLogger(__name__)


def fit_unbinned(
    func: Callable,
    data: np.ndarray,
    guess: np.ndarray = None,
    extended: bool = True,  # noqa: N803
    cost_func: str = "LL",
    simplex: bool = False,
    bounds: list[tuple[float, float], ...] = None,
    fixed: list[bool, ...] = None,
) -> tuple[np.ndarray, ...]:
    """Do a unbinned fit to data.
    Default is extended Log Likelihood fit, with option for other cost functions.

    Parameters
    ----------
    func
        the function to fit
    data
        the data values to be fit
    guess
        initial guess parameters
    extended
        run extended or non extended fit
    cost_func
        cost function to use. LL is ExtendedUnbinnedNLL, None is for just UnbinnedNLL
    simplex
        whether to include a round of simpson minimisation before main minimisation
    bounds
        Each tuple is (min, max) for the corresponding parameters.
        Bounds can be None, e.g. [(0,None), (0,10)]
    fixed
        list of parameter indices to fix

    Returns
    -------
    pars, errs, cov
        the best-fit parameters and their errors / covariance
    """
    if guess is None:
        raise NotImplementedError(
            "auto-guessing not yet implemented, you must supply a guess."
        )

    if cost_func == "LL":
        if extended is True:
            cost_func = cost.ExtendedUnbinnedNLL(data, func)

        else:
            cost_func = cost.UnbinnedNLL(data, func)
    if isinstance(guess, dict):
        m = Minuit(cost_func, **guess)
    else:
        m = Minuit(cost_func, *guess)
    if bounds is not None:
        if isinstance(bounds, dict):
            for arg, key in bounds.items():
                m.limits[arg] = key
        else:
            m.limits = bounds
    if fixed is not None:
        for fix in fixed:
            m.fixed[fix] = True
    if simplex is True:
        m.simplex().migrad()
    else:
        m.migrad()
    m.hesse()
    return m.values, m.errors, m.covariance



import numpy as np
from scipy.stats import expon

from pygama.math.functions.exponential import exponential
from pygama.math.functions.pygama_continuous import PygamaContinuous


def test_exponential_pdf():
    x = np.arange(1, 12)
    mu = 3.1
    sigma = 2.3
    lamb = 0.2

    par_array = [mu, sigma, lamb]

    y = exponential.pdf(x, *par_array)
    y_direct = exponential.get_pdf(x, *par_array)
    scipy_y = expon.pdf(x, mu, sigma / lamb)

    assert isinstance(exponential, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    x_lo = 0
    x_hi = 200

    y_sig, y_ext = exponential.pdf_ext(x, x_lo, x_hi, n_sig, *par_array)
    assert np.allclose(y_ext, n_sig * scipy_y, rtol=1e-8)
    assert np.allclose(y_sig, n_sig, rtol=1e-3)

    normalization = np.diff(expon.cdf(x[np.array([0, -1])], mu, sigma / lamb))
    y_norm = exponential.pdf_norm(x, x[0], x[-1], *par_array)

    assert np.allclose(y_norm, scipy_y / normalization, rtol=1e-8)


def test_exponential_cdf():
    x = np.arange(5, 12)
    mu = 1.1
    sigma = 2.4
    lamb = 0.3
    par_array = [mu, sigma, lamb]

    y = exponential.cdf(x, *par_array)
    y_direct = exponential.get_cdf(x, *par_array)
    scipy_y = expon.cdf(x, mu, sigma / lamb)

    assert isinstance(exponential, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    y_ext = exponential.cdf_ext(x, n_sig, *par_array)
    assert np.allclose(y_ext, n_sig * scipy_y, rtol=1e-8)

    normalization = np.diff(expon.cdf(x[np.array([0, -1])], mu, sigma / lamb))
    y_norm = exponential.cdf_norm(x, x[0], x[-1], *par_array)

    assert np.allclose(y_norm, scipy_y / normalization, rtol=1e-8)


def test_required_args():
    names = exponential.required_args()
    assert names[0] == "mu"
    assert names[1] == "sigma"
    assert names[2] == "lambda"


def test_name():
    assert exponential.name == "exponential"



from pathlib import Path

import numpy as np
import pytest
from lgdo import lh5

from pygama.evt import utils
from pygama.evt.modules import xtalk

config_dir = Path(__file__).parent / "configs"


@pytest.fixture(scope="module")
def files_config(lgnd_test_data, tmp_dir):
    tcm_path = "lh5/prod-ref-l200/generated/tier/tcm/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_tcm.lh5"
    outfile = f"{tmp_dir}/l200-p03-r001-phy-20230322T160139Z-tier_evt.lh5"

    return {
        "tcm": (lgnd_test_data.get_path(tcm_path), "hardware_tcm_1"),
        "dsp": (lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp")), "dsp", "ch{}"),
        "hit": (lgnd_test_data.get_path(tcm_path.replace("tcm", "hit")), "hit", "ch{}"),
        "evt": (outfile, "evt"),
    }


def test_xtalk_corrected_energy(lgnd_test_data, files_config):

    energy = np.array([[1, 2, 3], [4, 5, 6], [2, 0, 1], [0, 1, 0]])
    matrix = np.array([[0, 0, 1], [1, 0, 2], [0, 2, 0]])
    energy_corrected_zero_threshold = xtalk.xtalk_correct_energy_impl(
        energy, energy, matrix, None
    )

    assert np.all(
        energy_corrected_zero_threshold
        == (energy - np.array([[3, 7, 4], [6, 16, 10], [1, 4, 0], [0, 0, 2]]))
    )

    # test with a 2.1 threshold
    energy_corrected_two_threshold = xtalk.xtalk_correct_energy_impl(
        energy, energy, matrix, 2.1
    )
    assert np.all(
        energy_corrected_two_threshold
        == (energy - np.array([[3, 6, 0], [6, 16, 10], [0, 0, 0], [0, 0, 0]]))
    )


def test_gather_energy(lgnd_test_data, files_config):
    f = utils.make_files_config(files_config)
    tcm = utils.TCMData(
        table_key=lh5.read_as(f"/{f.tcm.group}/table_key", f.tcm.file, library="ak"),
        row_in_table=lh5.read_as(
            f"/{f.tcm.group}/row_in_table", f.tcm.file, library="ak"
        ),
    )
    energy = xtalk.gather_energy(
        "hit.cuspEmax_ctc_cal", tcm, f, np.array([1084803, 1084804])
    )
    n_rows = np.max(tcm.row_in_table) + 1
    assert isinstance(energy, np.ndarray)
    assert energy.ndim == 2
    assert np.shape(energy) == (n_rows, 2)


def test_filter_hits(lgnd_test_data, files_config):
    f = utils.make_files_config(files_config)
    tcm = utils.TCMData(
        table_key=lh5.read_as(f"/{f.tcm.group}/table_key", f.tcm.file, library="ak"),
        row_in_table=lh5.read_as(
            f"/{f.tcm.group}/row_in_table", f.tcm.file, library="ak"
        ),
    )
    n_rows = np.max(tcm.row_in_table) + 1

    filter = xtalk.filter_hits(
        f,
        tcm,
        "hit.cuspEmax_ctc_cal>5",
        np.zeros((n_rows, 2)),
        np.array([1084803, 1084804]),
    )

    assert isinstance(filter, np.ndarray)
    assert filter.ndim == 2
    assert np.shape(filter) == (n_rows, 2)



import numba as nb
import numpy as np
from numba import prange

from pygama.math.functions.pygama_continuous import PygamaContinuous
from pygama.utils import numba_math_defaults as nb_defaults
from pygama.utils import numba_math_defaults_kwargs as nb_kwargs


@nb.njit(**nb_kwargs)
def nb_linear_pdf(
    x: np.ndarray, x_lo: float, x_hi: float, m: float, b: float
) -> np.ndarray:
    r"""
    Normalised linear probability density function, w/ args: m, b. Its range of support is :math:`x\in(x_{lower},x_{upper})`.
    If :math:`x_{lower} = np.inf` and :math:`x_{upper} = np.inf`, then the function takes  :math:`x_{upper} = :func:`np.min(x)` and :math:`x_{upper} = :func:`np.amax(x)`
    It computes:


    .. math::
        pdf(x, x_{lower}, x_{upper}, m, b) = \frac{mx+b}{\frac{m}{2}(x_{upper}^2-x_{lower}^2)+b(x_{upper}-x_{lower})}


    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    x_lo
        The lower bound of the distribution
    x_hi
        The upper bound of the distribution
    m
        The slope of the linear part
    b
        The y-intercept of the linear part
    """
    norm = (m / 2) * (x_hi**2 - x_lo**2) + b * (x_hi - x_lo)

    result = np.empty_like(x, np.float64)
    for i in prange(x.shape[0]):
        result[i] = (m * x[i] + b) / norm
    return result


@nb.njit(**nb_kwargs)
def nb_linear_cdf(
    x: np.ndarray, x_lo: float, x_hi: float, m: float, b: float
) -> np.ndarray:
    r"""
    Normalised linear cumulative density function, w/ args: m, b. Its range of support is :math:`x\in(x_{lower},x_{upper})`.
    If :math:`x_{lower} = np.inf` and :math:`x_{upper} = np.inf`, then the function takes  :math:`x_{upper} = :func:`np.min(x)` and :math:`x_{upper} = :func:`np.amax(x)`
    It computes:


    .. math::
        cdf(x, x_{lower}, x_{upper}, m, b) = \frac{\frac{m}{2}(x^2-x_{lower}^2)+b(x-x_{lower})}{\frac{m}{2}(x_{upper}^2-x_{lower}^2)+b(x_{upper}-x_{lower})}


    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    x_lo
        The lower bound of the distribution
    x_hi
        The upper bound of the distribution
    m
        The slope of the linear part
    b
        The y-intercept of the linear part
    """
    norm = (m / 2) * (x_hi**2 - x_lo**2) + b * (x_hi - x_lo)

    result = np.empty_like(x, np.float64)
    for i in prange(x.shape[0]):
        result[i] = (m / 2 * (x[i] ** 2 - x_lo**2) + b * (x[i] - x_lo)) / norm
    return result


@nb.njit(**nb_defaults(parallel=False))
def nb_linear_scaled_pdf(
    x: np.ndarray, x_lo: float, x_hi: float, area: float, m: float, b: float
) -> np.ndarray:
    r"""
    Scaled linear probability distribution, w/ args: m, b.
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    x_lo
        The lower bound of the distribution
    x_hi
        The upper bound of the distribution
    area
        The number of counts in the signal
    m
        The slope of the linear part
    b
        The y-intercept of the linear part
    """

    return area * nb_linear_pdf(x, x_lo, x_hi, m, b)


@nb.njit(**nb_defaults(parallel=False))
def nb_linear_scaled_cdf(
    x: np.ndarray, x_lo: float, x_hi: float, area: float, m: float, b: float
) -> np.ndarray:
    r"""
    Linear cdf scaled by the area for extended binned fits
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    x_lo
        The lower bound of the distribution
    x_hi
        The upper bound of the distribution
    area
        The number of counts in the signal
    m
        The slope of the linear part
    b
        The y-intercept of the linear part
    """

    return area * nb_linear_cdf(x, x_lo, x_hi, m, b)


class LinearGen(PygamaContinuous):

    def __init__(self, *args, **kwargs):
        self.x_lo = None
        self.x_hi = None
        super().__init__(*args, **kwargs)

    def _argcheck(self, x_lo, x_hi, m, b):
        return True

    def _pdf(self, x: np.ndarray, x_lo: float, x_hi: float, m, b) -> np.ndarray:
        x.flags.writeable = True
        return nb_linear_pdf(x, x_lo[0], x_hi[0], m[0], b[0])

    def _cdf(self, x: np.ndarray, x_lo: float, x_hi: float, m, b) -> np.ndarray:
        x.flags.writeable = True
        return nb_linear_cdf(x, x_lo[0], x_hi[0], m[0], b[0])

    def get_pdf(
        self, x: np.ndarray, x_lo: float, x_hi: float, m: float, b: float
    ) -> np.ndarray:
        return nb_linear_pdf(x, x_lo, x_hi, m, b)

    def get_cdf(
        self, x: np.ndarray, x_lo: float, x_hi: float, m: float, b: float
    ) -> np.ndarray:
        return nb_linear_cdf(x, x_lo, x_hi, m, b)

    # Because this function is already normalized over its limited support, we need to alias get_pdf as pdf_norm
    def pdf_norm(
        self, x: np.ndarray, x_lo: float, x_hi: float, m: float, b: float
    ) -> np.ndarray:
        return nb_linear_pdf(x, x_lo, x_hi, m, b)

    def cdf_norm(
        self, x: np.ndarray, x_lo: float, x_hi: float, m: float, b: float
    ) -> np.ndarray:
        return nb_linear_cdf(x, x_lo, x_hi, m, b)

    def pdf_ext(
        self, x: np.ndarray, x_lo: float, x_hi: float, area: float, m: float, b: float
    ) -> np.ndarray:
        return np.diff(
            nb_linear_scaled_cdf(np.array([x_lo, x_hi]), x_lo, x_hi, area, m, b)
        )[0], nb_linear_scaled_pdf(x, x_lo, x_hi, area, m, b)

    def cdf_ext(
        self, x: np.ndarray, x_lo: float, x_hi: float, area: float, m: float, b: float
    ) -> np.ndarray:
        return nb_linear_scaled_cdf(x, x_lo, x_hi, area, m, b)

    def required_args(self) -> tuple[str, str, str, str]:
        return "x_lo", "x_hi", "m", "b"


linear = LinearGen(name="linear")
